/* $Id: e2_cache.c,v 1.11 2004/06/04 15:37:58 cairo Exp $
 * 
 * Copyright (C) 2004 Florian Zaehringer (flo.zaehringer@web.de)
 * Copyright (C) 2004 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file COPYING. If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * @file e2_cache.c
 * @brief cache system
 *
 * This file contains functions for e2's cache system: eg. to register and
 * unregister caches, and to write and read them to/from disk.
 **/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <gtk/gtk.h>
#include <glib/gprintf.h>
#include "e2_cache.h"

//external dependencies
#include "e2_cl_option.h"
#include "debug.h"
#include "build.h"

static GList *cache_list;
static GList *unknown_list;

  /*****************/
 /***** utils *****/
/*****************/

static gint str_compare (E2_Cache *cache_element, gchar *name)
{
	return !g_str_equal (name, cache_element->name);
}

  /******************/
 /***** public *****/
/******************/

void e2_cache_int_register (gchar *name, gint *dbl, gint def)
{
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (WARN, "trying to register double cache '%s' more than one time", name);
		return;
	}
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	
	E2_Cache *cache = malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_INT;
	if (found == NULL)
	{
		printd (DEBUG, "cache '%s' not found", name);
		cache->name = g_strdup (name);
		*dbl = def;
	} else
	{
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		gchar *end = NULL;
		gchar *str = (gchar *) unknown->data;
		*dbl = g_ascii_strtoull (str, &end, 10);
		if (end == str)
		{
			printd (WARN, "cache '%s' data conversion failed from '%s", name, str);
			*dbl = def;
		} else
			g_free (str);
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = GINT_TO_POINTER (dbl);
	cache_list = g_list_append (cache_list, cache);
}

void e2_cache_double_register (gchar *name, gdouble *dbl, gdouble def)
{
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (WARN, "trying to register double cache '%s' more than one time", name);
		return;
	}
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	
	E2_Cache *cache = malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_DOUBLE;
	if (found == NULL)
	{
		printd (DEBUG, "cache '%s' not found", name);
		cache->name = g_strdup (name);
		*dbl = def;
	} else
	{
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		gchar *end = NULL;
		gchar *str = (gchar *) unknown->data;
		*dbl = g_ascii_strtod (str, &end);
		if (end == str)
		{
			printd (WARN, "cache '%s' data conversion failed from '%s", name, str);
			*dbl = def;
		} else
			g_free (str);
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer *) dbl;
	cache_list = g_list_append (cache_list, cache);
}

void e2_cache_str_register (gchar *name, gchar **str)
{
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (WARN, "trying to register str cache '%s' more than one time", name);
		return;
	}
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	
	E2_Cache *cache = malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_STR;
	if (found == NULL)
	{
		printd (DEBUG, "cache '%s' not found", name);
		cache->name = g_strdup (name);
		*str = NULL;
	} else
	{
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		*str = (gchar *) unknown->data;
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer *) str;
	cache_list = g_list_append (cache_list, cache);
}

E2_Cache *e2_cache_list_register (gchar *name, GList **list)
{
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (WARN, "trying to register list cache '%s' more than one time", name);
		return NULL;
	}
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	
	E2_Cache *cache = malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_LIST;
	cache->sync_func = NULL;
	cache->sync_data = NULL;
	if (found == NULL)
	{
		printd (DEBUG, "cache '%s' not found", name);
		cache->name = g_strdup (name);
		*list = NULL;
	} else
	{
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		*list = (GList *) unknown->data;
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer *) list;
	cache_list = g_list_append (cache_list, cache);
	return cache;
}

void e2_cache_list_unregister (gchar *name)
{
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{
		printd (WARN, "trying to unregister cache '%s' which doesn't exist", name);
		return;
	}
	cache_list = g_list_remove_link (cache_list, found);
	E2_Cache *cache = found->data;
	E2_Cache *unknown = malloc (sizeof (E2_Cache));
	unknown->name = cache->name;
	unknown->type = cache->type;
	unknown->data = *cache->data;
	unknown_list = g_list_append (unknown_list, unknown);
	g_free (cache);
}

void e2_cache_read_from_string (gchar *f[])
{
	gboolean list_mode = FALSE;
	GList *list = NULL;
	E2_Cache *cache = NULL;
	
	gint i = 0;
	gchar *line;
	
	while ((line = f[i++]) != NULL)
	{
		g_strchomp (line);
		//ignore empty lines and comments
		if (*line == '\0')
			continue;
		if (line[0] == '#')
			continue;
		
		if (line[0] == '<')
		{
			list_mode = TRUE;
			list = NULL;
			cache = malloc (sizeof (E2_Cache));
			cache->name = g_strdup (line + 1);
			cache->type = E2_CACHE_TYPE_LIST;
			continue;
		}
		
		if (line[0] == '>')
		{
			list_mode = FALSE;
			if (list == NULL)
				continue;
			cache->data = (gpointer *) list;
			unknown_list = g_list_append (unknown_list, cache);
			continue;
		}
		
		if (list_mode)
		{
			list = g_list_append (list, g_strdup (line));
		} else
		{
			cache = malloc (sizeof (E2_Cache));
			gchar **split = g_strsplit (line, "=", 2);
			cache->name = g_strdup (split[0]);
			cache->data = (gpointer *) g_strdup (split[1]);
			cache->type = E2_CACHE_TYPE_STR;
			g_strfreev (split);
			unknown_list = g_list_append (unknown_list, cache);
		}
	} 
}

void e2_cache_read_from_file (gchar *fn)
{
	//find absolute path to config file
	gchar *filename = g_build_filename (e2_cl_options.config_dir, fn, NULL);
	printd (DEBUG, "cache read from file '%s'", filename);
	gchar *cont;
	if ((g_file_get_contents (filename, &cont, NULL, NULL)) == FALSE)
	{
		printd (WARN, "could not open cache file '%s'", filename);
	} else
	{
		gchar **split = g_strsplit (cont, "\n", -1);
		e2_cache_read_from_string (split);
		g_strfreev (split);
		g_free (cont);
	}
	g_free (filename);
}

void e2_cache_write_to_file ()
{
	if (cache_list == NULL)
	{
		printd (DEBUG, "no caches registered");
		return;
	}
	gchar *filename;
	filename = g_build_filename (e2_cl_options.config_dir, DEFAULT_CACHE_FILE, NULL);
	printd (DEBUG, "write cache file: %s", filename);

	FILE *f;
	if ((f = fopen (filename, "w")) == NULL)
	{
		printd (WARN, "cannot write to cache file: %s (%s)", filename, g_strerror (errno));
		return;
	}
	g_free (filename);

	g_fprintf 
	(
		f, "# " PROGNAME " (v" VERSION ")\n\n"
		"# this is the " PROGNAME " cache file.\n"
		"# this is a generated file and will be overwritten automatically each run!\n\n"
	);

	GList *list;
	for (list = cache_list; list != NULL; list = g_list_next (list))
	{
		E2_Cache *cache = list->data;
		switch (cache->type)
		{
			case E2_CACHE_TYPE_BOOL:
				g_fprintf (f, "%s=%s\n", cache->name, *cache->data ? "true" : "false");
				break;
			case E2_CACHE_TYPE_DOUBLE:
			{
				gchar doubl[G_ASCII_DTOSTR_BUF_SIZE];
				g_ascii_dtostr (doubl, G_ASCII_DTOSTR_BUF_SIZE, *((gdouble *) cache->data));
				g_fprintf (f, "%s=%s\n", cache->name, doubl);
			}
				break;
			case E2_CACHE_TYPE_INT:
				g_fprintf (f, "%s=%d\n", cache->name, GPOINTER_TO_INT (*cache->data));
				break;
			case E2_CACHE_TYPE_STR:
				g_fprintf (f, "%s=%s\n", cache->name, (gchar*) *cache->data);
				break;
			case E2_CACHE_TYPE_LIST:
				if (cache->sync_func != NULL)
				{
					void (*fun) (gpointer) = cache->sync_func;
					fun (cache->sync_data);
				}
				if ((cache->data != NULL) && (*cache->data != NULL))
				{
					g_fprintf (f, "<%s\n", cache->name);
					GList *tmp;
					for (tmp = (GList *) *cache->data; tmp != NULL; tmp = g_list_next (tmp))
					{
						g_fprintf (f, "%s\n", (gchar *) tmp->data);
					}
					g_fprintf (f, ">\n");
				}
				break;
			default:
				printd (WARN, "don't know how to write '%s' to cache file",
					cache->name);
				break;
		}
	}
	//FIXME: write unknown cache options to file
	fclose (f);
}

/**
 * @brief initialize the cache
 *
 * This function will initilize the cache system of e2. It will automatically
 * be called at startup.
 *
 * @return 
 **/
void e2_cache_init ()
{
	cache_list = NULL;
	unknown_list = NULL;
	e2_cache_read_from_file (DEFAULT_CACHE_FILE);
}
