/* $Id: e2_option.c,v 1.26 2004/05/08 13:54:25 cairo Exp $
 *
 * Copyright (C) 2003-2004 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file COPYING. If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * @file e2_option.c
 * @brief general option handling
 *
 * This file contains general functions to register, unregister and set
 * options.
 **/

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include "emelfm2.h"
#include "e2_option.h"
#include "e2_option_tree.h"
#include "debug.h"
#include "build.h"
#include "e2_action.h"
#include "e2_cl_option.h"

//to save option references
GHashTable *options_hash;
GHashTable *options_queue;
GPtrArray *options_array;

  /******************/
 /***** public *****/
/******************/

/**
 * @brief check if config dir exists
 *
 * initialize config directory pointer and check if it really exists. if not,
 * create it or print an appropriate error message through the debug system.
 * 
 * @return 
 **/
static void check_config_dir ()
{
	//check if config dir is already set through a command line option
	if (!((e2_cl_options.config_dir != NULL) && (*e2_cl_options.config_dir != '\0')))
	{
		//if not, build it
		const gchar *home = g_get_home_dir ();
		if (home != NULL)
			e2_cl_options.config_dir = g_build_filename (home, ".emelfm2", NULL);
		else
		{
			printd (ERROR, "could not find out home directory. "
				"please set the HOME environment variable!");
			exit (1);
		}
	} else
	{
		//if the command line config dir path is not absolute,
		//correct this to prevent problems later on
		if (!g_path_is_absolute (e2_cl_options.config_dir))
		{
			gchar *current_dir = g_get_current_dir ();
			gchar *config_dir = g_build_filename (current_dir, e2_cl_options.config_dir, NULL);
			g_free (current_dir);
			g_free (e2_cl_options.config_dir);
			e2_cl_options.config_dir = config_dir;
		}
	}
	//check if the config dir is really there,
	//if not, create it
	if ((g_file_test (e2_cl_options.config_dir, G_FILE_TEST_EXISTS)) != TRUE)
	{
		printd (INFO, "creating config directory '%s'", e2_cl_options.config_dir);
		if (mkdir (e2_cl_options.config_dir, 0777))
		{
			printd (WARN, "could not create config directory '%s'",
				e2_cl_options.config_dir);
		}
	} else if ((g_file_test (e2_cl_options.config_dir, G_FILE_TEST_IS_DIR)) != TRUE)
	{
		printd (WARN, "could not create config directory '%s', "
			"something is in the way", e2_cl_options.config_dir);
	}
}

/**
 * @brief initialize option system
 *
 * internal data structures of the option system will be initialized. 
 * this includes creation of a hash table for fast option lookup.
 * this function is called only once on startup by the main function
 *
 * @return 
 **/
void e2_option_init ()
{
	check_config_dir ();

	//save config dir timestamp
	struct stat stat_buf;
	if (stat (e2_cl_options.config_dir, &stat_buf) == 0)
		cfg.config_mtime = stat_buf.st_mtime;

	options_hash = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
	options_queue = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
	options_array = g_ptr_array_new ();

	e2_option_default_register ();
}

/**
 * @brief read config options
 *
 * @param f NULL-terminated array of strings in config file format
 *
 * read config options from @a f and parse them. this is usualy only
 * done at startup.
 *
 * @return 
 **/
void e2_option_read_from_string (gchar *f[])
{
	//save if a tree option was found because they are several lines long
	gboolean tree_mode = FALSE;
	//helper GList for tree options used to process parent/child dependencies
	GList *parents = g_list_append (NULL, NULL);
	//line counter
	gint i = 0;
	//pointer to the current line
	gchar *line;
	
	//helper for tree option processing to save pointer to the tree option
	E2_OptionSet *set = NULL;
		
	while ((line = f[i++]) != NULL)
	{
		g_strchomp (line);
		//ignore empty lines and comments
		if (*line == '\0') continue;
		if (line[0] == '#') continue;
		
		//tree option begins
		if (line[0] == '<')
		{
			tree_mode = TRUE;
			//try to find pointer to option
			set = _e2_option_get (line + 1);
			//found config for this tree option, so don't init it with the
			//default config later on
			if (set != NULL)
				set->ex.tree.found_in_config_file = TRUE;
			else
				//unknown tree option, save for later processing
				//(when the option is actually added)
				i = e2_option_tree_register_unknown (line + 1, i - 1, f);
			
			continue;
		}
		
		//tree option ends
		if (line[0] == '>')
		{
			tree_mode = FALSE;
			continue;
		}
		
		if (tree_mode)
		{
			//be on the save side for non-existing list options
			if (set == NULL)
				continue;

			//depth of the current tree option row
			gint level = 0;
			while (line[level] == '\t')
				level++;
			
			//get rid of the tabs at the start of the line
			line += level;

			//"unescape" the first character in case it is escaped
			if ((line[0] == '\\') && ((line[1] == '<') || (line[1] == '>')))
				line++;

			//FIXME: strsplit isn't the best coice anymore because
			//that makes unescaping too complicated
			gchar **split = g_strsplit (line, "|", -1);
			void *tmp[set->ex.tree.columns_num];
			//split counter, increased when stepping through **split
			gint i = 0;
			//tree column counter
			gint j = 0;
			//tree column pointer
			GList *column;
			//unescape helper to be able to unescaped concatted
			//values later on
			GList *free = NULL;
			//transform splitted values to void pointers and
			//add default values in case of a missing value/field
			for (column = set->ex.tree.columns; column != NULL; column = g_list_next (column))
			{
				//current tree column pointer
				E2_OptionTreeSet *opt = column->data;
				if (split[i] != NULL)
				{
					gchar *value = split[i];
					//unescape | and concat values
					while (split[i + 1] != '\0')
					{
						gint len = strlen (value);
						if ((len > 0) && (value[len - 1] == '\\'))
						{
							value[len - 1] = '\0';
							value = g_strconcat (value, "|", split[i + 1], NULL);
							//save pointer to free it after
							//knowing how long this really gets
							//(there could be several escaped | in one line)
							free = g_list_append (free, value);
							i++;
						} else
							break;
					}
					//get void pointer to add it to the model
					tmp[j] = _e2_option_tree_get_void (opt, value);
				} else
					//use default value
					tmp[j] = _e2_option_tree_get_void (opt, opt->sdef);
				j++;
				i++;
			}

			GtkTreeIter iter;
			//add tree option row
			_e2_option_tree_add (set, &iter, g_list_nth_data (parents, level),
				FALSE, TRUE, set->ex.tree.columns_num, tmp);
			//add iter pointer to be able to add a child to this row later on
			GList *list = g_list_nth (parents, level + 1);
			if (list == NULL)
				parents = g_list_append (parents, gtk_tree_iter_copy (&iter));
			else
			{
				gtk_tree_iter_free (list->data);
				list->data = gtk_tree_iter_copy (&iter);
			}

			//cleanup
			g_strfreev (split);
			g_list_foreach (free, (GFunc) g_free, NULL);
			g_list_free (free);
		} else
		//normal option (not a tree option)
		{
			gchar **split = g_strsplit (line, "=", 2);
			if ((e2_option_set_str (split[0], split[1])) == FALSE)
				printd (WARN, "could not read option '%s' from config file", split[0]);
			g_strfreev (split);
		}
	}

	//cleanup
	parents = g_list_remove (parents, NULL);
	g_list_foreach (parents, (GFunc) gtk_tree_iter_free, NULL);
	g_list_free (parents);
}

gboolean e2_option_register (E2_OptionSet *set)
{
	set->widget = NULL;
	set->should_be_freed = FALSE;

	if ((g_hash_table_lookup (options_hash, set)) == NULL)
	{
		g_ptr_array_add (options_array, set);
		g_hash_table_insert (options_hash, set->name, set);
		g_hook_list_init (&set->hook_value_changed, sizeof (GHook));
		set->hook_freezed = FALSE;
		gchar *value;
		if ((value = g_hash_table_lookup (options_queue,
					set->name)) != NULL)
		{
			if (set->type == E2_OPTION_TYPE_TREE)
				set->ex.tree.unknown = (gchar **) value;
			else
			{
				_e2_option_set_str (set, value);
				g_free (value);
			}
			g_hash_table_remove (options_queue, set->name);
			return TRUE;
		}
	}
	return FALSE;
}

/**
 * @brief get E2_OptionSet with warning
 *
 * @param option name of the option
 *
 * @return option or NULL in case i cannot be found
 **/
E2_OptionSet *e2_option_get (gchar *option)
{
	E2_OptionSet *set;
	if ((set = (E2_OptionSet *) g_hash_table_lookup (options_hash, option)) == NULL)
	{
		printd (WARN, "trying to get option '%s' which doesn't exist", option);
		return NULL;
	}
	return set;
}

/**
 * @brief get E2_OptionSet without warning
 *
 * @param option name of the option
 *
 * @return option or NULL in case i cannot be found
 **/
E2_OptionSet *_e2_option_get (gchar *option)
{
	return (E2_OptionSet *) g_hash_table_lookup (options_hash, option);
}

void *_e2_option_void_get (E2_OptionSet *set, gchar *option)
{
	switch (set->type)
	{
		case E2_OPTION_TYPE_BOOL:
			if (g_str_equal (option, "true")) return (void *) TRUE;
			else return (void *) FALSE;
			break;
		case E2_OPTION_TYPE_STR:
		case E2_OPTION_TYPE_FONT:
		case E2_OPTION_TYPE_ICON:
		case E2_OPTION_TYPE_SEL:
			return (void *) option;
		default:
			return NULL;
			break;
	}
	return NULL;
}

void e2_option_clear ()
{
	gint i;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
		if (set->type == E2_OPTION_TYPE_TREE)
			gtk_tree_store_clear (set->ex.tree.model);
	}
}

void e2_option_attach_value_changed (gchar *option, GtkWidget *widget,
	gpointer func, gpointer data)
{
	E2_OptionSet *set = e2_option_get (option);
	if (set != NULL)
		_e2_option_attach_value_changed (set, widget, func, data);
}


void _e2_option_attach_value_changed (E2_OptionSet *set, GtkWidget *widget,
	gpointer func, gpointer data)
{
	GHook *hook = e2_hook_register (&set->hook_value_changed, func, data);
	E2_Duo *duo = g_malloc (sizeof (E2_Duo));
	duo->a = hook;
	duo->b = &set->hook_value_changed;
	g_object_set_data_full (G_OBJECT (widget), "e2-option-attach-value-changed",
		duo, (GDestroyNotify) e2_hook_unattach_cb);
}

void destroy_config ()
{
	g_hash_table_destroy (options_hash);
	g_hash_table_destroy (options_queue);
	g_ptr_array_free (options_array, TRUE);
}

