/*
 * $Id: e2_list.c,v 1.9 2004/05/08 13:54:25 cairo Exp $
 *
 * Copyright (C) 2004 tooar <tooar@gmx.net>
 * Copyright (C) 1999 Michael Clark.
 *
 * emelFM2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/**
 * @file e2_list.c 
 * @brief glist utilities
 *
 * glist utility-functions
 *
 **/

#include <stdlib.h>
#include <string.h>
#include "emelfm2.h"

/**
 * @brief compare function for glists with strings as their data
 * 
 * compare function for glists with strings as their data
 *
 * @param a  first string
 * @param b  second string
 *
 * @return 0 if the strings are equal, 1 otherwise
 **/
gint e2_list_strcmp (gchar *a, gchar *b)
{
	return !g_str_equal (a, b);
}

void e2_list_update_history (const gchar *command, GList **history, gint *cur,
	gint max, gboolean doubl)
{
	GList *tmp = doubl ? NULL : g_list_find_custom (*history, command,
		(GCompareFunc) e2_list_strcmp);

	//found the command in history
	if (tmp != NULL)
	{
		*history = g_list_remove_link (*history, tmp);
		*history = g_list_append (*history, g_strdup (command));
		g_free (tmp->data);
		g_list_free_1 (tmp);
	} else
	//didn't find the command in history
	{
		if ((*history != NULL) && (g_str_equal (command,
			g_list_nth_data (*history, g_list_length (*history) - 1))))
			return;
		*history = g_list_append (*history, g_strdup (command));
		if (g_list_length (*history) > max)
		{
			tmp = g_list_nth (*history, 0);
			*history = g_list_remove_link (*history, tmp);
			g_free (tmp->data);
			g_list_free_1 (tmp);
		}
	}
	if (cur != NULL)
		*cur = g_list_length (*history);
}

/**
 *
 * @brief convert a GList of strings to an array of strings
 *
 * convert a GList of strings to an NULL-terminated array of strings
 *
 * @param list GList with gchar pointers as data
 *
 * @return NULL-terminated array of strings from @a list
 **/
gchar **e2_list_to_strv (GList *list)
{
	gint len = g_list_length (list);
	gchar **strv = malloc ((len + 1) * sizeof (gpointer));
	gint i;
	GList *l = list;
	for (i = 0; i < len; i++)
	{
		strv[i] = l->data;
		l = g_list_next (l);
	}
	strv[len] = NULL;
	return strv;
}

/**
 * @brief copy a GList of strings
 *
 * This function copies a GList and also copies the data of each element
 * with g_strdup.
 *
 * @param list the GList to copy
 *
 * @return new GList that has to be freed
 **/
GList *e2_list_copy_with_data (GList *list)
{
	GList *copy = NULL;
	GList *tmp;
	for (tmp = g_list_first (list); tmp != NULL; tmp = g_list_next (tmp))
	{
		copy = g_list_append (copy, g_strdup (tmp->data));
	}
	return copy;
}

void e2_list_free_with_data (GList **list)
{
	if ((list != NULL) && (*list != NULL))
	{
		g_list_foreach (*list, (GFunc) g_free, NULL);
		g_list_free (*list);
		*list = NULL;
	}
}

void free_clist_data (GtkWidget *clist)
{
	gint i;

	for (i = 0; i < GTK_CLIST (clist)->rows; i++)
		g_free (gtk_clist_get_row_data (GTK_CLIST (clist), i));
}

GList *
clist_data_to_glist (GtkWidget *clist)
{
	gint i;
	GList *list = NULL;

	for (i = 0; i < GTK_CLIST (clist)->rows; i++)
		list = g_list_append (list, gtk_clist_get_row_data (GTK_CLIST (clist), i));
	return list;
}

GList *
string_glist_find (GList *list, gchar *search_text)
{
	GList *tmp;
	gchar *curr;

	for (tmp = list; tmp != NULL; tmp = tmp->next)
	{
		curr = tmp->data;
		if (g_str_equal (curr, search_text)) return tmp;
	}
	return NULL;
}

void
clist_select_rows (GtkWidget *clist, GList *rows)
{
	GList *tmp;
	gint row;

	if (rows == NULL)
	{
		gtk_clist_select_row (GTK_CLIST (clist), 0, 0);
		GTK_CLIST (clist)->focus_row = 0;
	}
	else
	{
		for (tmp = rows; tmp != NULL; tmp = tmp->next)
		{
			row = (gint)tmp->data;
			if (row < GTK_CLIST (clist)->rows)
			{
				gtk_clist_select_row (GTK_CLIST (clist), row, 0);
				GTK_CLIST (clist)->focus_row = row;
			}
			else
			{
				gtk_clist_select_row (GTK_CLIST (clist), GTK_CLIST (clist)->rows-1, 0);
				GTK_CLIST (clist)->focus_row = GTK_CLIST (clist)->rows-1;
			}
		} 
	}
	gtk_widget_grab_focus (clist);
}

gboolean
clist_row_is_selected (GtkWidget *clist, gint row)
{
	GList *tmp;

	for (tmp = GTK_CLIST (clist)->selection; tmp != NULL; tmp = tmp->next)
		if (row == (gint)tmp->data)
			return TRUE;
	return FALSE;
}
